quietflamingo commited on
Commit
8061ca4
·
verified ·
1 Parent(s): e46c57b

Delete flash_attn_triton.py

Browse files
Files changed (1) hide show
  1. flash_attn_triton.py +0 -1112
flash_attn_triton.py DELETED
@@ -1,1112 +0,0 @@
1
- # Copyright 2022 MosaicML Examples authors
2
- # SPDX-License-Identifier: Apache-2.0
3
-
4
- """Triton implementation of Flash Attention.
5
-
6
- # Copyright (c) 2022, Tri Dao.
7
- #
8
- # Licensed under the Apache License, Version 2.0 (the "License");
9
- # you may not use this file except in compliance with the License.
10
- # You may obtain a copy of the License at
11
- #
12
- # http://www.apache.org/licenses/LICENSE-2.0
13
- #
14
- # Unless required by applicable law or agreed to in writing, software
15
- # distributed under the License is distributed on an "AS IS" BASIS,
16
- # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
17
- # See the License for the specific language governing permissions and
18
- # limitations under the License.
19
-
20
- *Experimental* implementation of FlashAttention in Triton.
21
- We use the FlashAttention implementation from Phil Tillet a starting point.
22
- https://github.com/openai/triton/blob/master/python/tutorials/06-fused-attention.py
23
-
24
- Changes:
25
- - Implement both causal and non-causal attention.
26
- - Implement both self-attention and cross-attention.
27
- - Support arbitrary seqlens (not just multiples of 128), for both forward and backward.
28
- - Support all head dimensions up to 128 (not just 16, 32, 64, 128), for both forward and backward.
29
- - Support attention bias.
30
- - Speed up the forward pass a bit, and only store the LSE instead of m and l.
31
- - Make the backward for d=128 much faster by reducing register spilling.
32
- - Optionally parallelize the backward pass across seqlen_k, to deal with the case of
33
- small batch size * nheads.
34
-
35
- Caution:
36
- - If you plan to use headdim other than 64 and 128, you should test for race conditions
37
- (due to the Triton compiler), as done in tests/test_flash_attn.py
38
- "test_flash_attn_triton_race_condition". I've tested and fixed many race conditions
39
- for different head dimensions (40, 48, 64, 128, 80, 88, 96), but I'm still not 100% confident
40
- that there are none left for other head dimensions.
41
- Differences between this Triton version and the CUDA version:
42
- - Triton version doesn't support dropout.
43
- - Triton forward is generally faster than CUDA forward.
44
- - Triton backward is faster than CUDA backward when batch * nheads is small, and when headdim=64.
45
- It is slightly slower when headdim=128 and batch * nheads is large.
46
- - Triton version doesn't yet support different sequence lengths in a batch (i.e., RaggedTensor/NestedTensor).
47
- """
48
-
49
- import math
50
-
51
- import torch
52
- import triton # type: ignore (reportMissingImports)
53
- import triton.language as tl # type: ignore (reportMissingImports)
54
- from einops import repeat
55
-
56
-
57
- @triton.autotune(
58
- configs=[
59
- triton.Config({
60
- 'BLOCK_M': 128,
61
- 'BLOCK_N': 128
62
- },
63
- num_warps=8,
64
- num_stages=1),
65
- # This config has a race condition when EVEN_M == False, disabling it for now.
66
- # triton.Config({"BLOCK_M": 64, "BLOCK_N": 64}, num_warps=4, num_stages=1),
67
- ],
68
- key=[
69
- 'CACHE_KEY_SEQLEN_Q', 'CACHE_KEY_SEQLEN_K', 'BIAS_TYPE', 'IS_CAUSAL',
70
- 'BLOCK_HEADDIM'
71
- ])
72
- @triton.heuristics({
73
- 'EVEN_M': lambda args: args['seqlen_q'] % args['BLOCK_M'] == 0,
74
- 'EVEN_N': lambda args: args['seqlen_k'] % args['BLOCK_N'] == 0,
75
- 'EVEN_HEADDIM': lambda args: args['headdim'] == args['BLOCK_HEADDIM'],
76
- })
77
- @triton.jit
78
- def _fwd_kernel(
79
- Q,
80
- K,
81
- V,
82
- Bias,
83
- Out,
84
- Lse,
85
- TMP, # NOTE: TMP is a scratchpad buffer to workaround a compiler bug
86
- softmax_scale,
87
- stride_qb,
88
- stride_qh,
89
- stride_qm,
90
- stride_kb,
91
- stride_kh,
92
- stride_kn,
93
- stride_vb,
94
- stride_vh,
95
- stride_vn,
96
- stride_bb,
97
- stride_bh,
98
- stride_bm,
99
- stride_ob,
100
- stride_oh,
101
- stride_om,
102
- nheads,
103
- seqlen_q,
104
- seqlen_k,
105
- seqlen_q_rounded,
106
- headdim,
107
- CACHE_KEY_SEQLEN_Q,
108
- CACHE_KEY_SEQLEN_K,
109
- BIAS_TYPE: tl.constexpr,
110
- IS_CAUSAL: tl.constexpr,
111
- BLOCK_HEADDIM: tl.constexpr,
112
- EVEN_M: tl.constexpr,
113
- EVEN_N: tl.constexpr,
114
- EVEN_HEADDIM: tl.constexpr,
115
- BLOCK_M: tl.constexpr,
116
- BLOCK_N: tl.constexpr,
117
- ):
118
- start_m = tl.program_id(0)
119
- off_hb = tl.program_id(1)
120
- off_b = off_hb // nheads
121
- off_h = off_hb % nheads
122
- # off_b = tl.program_id(1)
123
- # off_h = tl.program_id(2)
124
- # off_hb = off_b * nheads + off_h
125
- # initialize offsets
126
- offs_m = start_m * BLOCK_M + tl.arange(0, BLOCK_M)
127
- offs_n = tl.arange(0, BLOCK_N)
128
- offs_d = tl.arange(0, BLOCK_HEADDIM)
129
- # Initialize pointers to Q, K, V
130
- # Adding parenthesis around indexing might use int32 math instead of int64 math?
131
- # https://github.com/openai/triton/issues/741
132
- # I'm seeing a tiny bit of difference (5-7us)
133
- q_ptrs = Q + off_b * stride_qb + off_h * stride_qh + (
134
- offs_m[:, None] * stride_qm + offs_d[None, :])
135
- k_ptrs = K + off_b * stride_kb + off_h * stride_kh + (
136
- offs_n[:, None] * stride_kn + offs_d[None, :])
137
- v_ptrs = V + off_b * stride_vb + off_h * stride_vh + (
138
- offs_n[:, None] * stride_vn + offs_d[None, :])
139
- if BIAS_TYPE == 'vector':
140
- b_ptrs = Bias + off_b * stride_bb + off_h * stride_bh + offs_n
141
- elif BIAS_TYPE == 'matrix':
142
- b_ptrs = Bias + off_b * stride_bb + off_h * stride_bh + (
143
- offs_m[:, None] * stride_bm + offs_n[None, :])
144
- else:
145
- raise ValueError("BIAS_TYPE must be one of {'vector', 'matrix'}")
146
- # initialize pointer to m and l
147
- t_ptrs = TMP + off_hb * seqlen_q_rounded + offs_m
148
- lse_i = tl.zeros([BLOCK_M], dtype=tl.float32) - float('inf')
149
- m_i = tl.zeros([BLOCK_M], dtype=tl.float32) - float('inf')
150
- acc_o = tl.zeros([BLOCK_M, BLOCK_HEADDIM], dtype=tl.float32)
151
- # load q: it will stay in SRAM throughout
152
- # [2022-10-30] TD: Triton bug - in the case of EVEN_M=True and EVEN_N=False, if we just call
153
- # tl.load(q_ptrs), we get the wrong output!
154
- if EVEN_M & EVEN_N:
155
- if EVEN_HEADDIM:
156
- q = tl.load(q_ptrs)
157
- else:
158
- q = tl.load(q_ptrs, mask=offs_d[None, :] < headdim, other=0.0)
159
- else:
160
- if EVEN_HEADDIM:
161
- q = tl.load(q_ptrs, mask=offs_m[:, None] < seqlen_q, other=0.0)
162
- else:
163
- q = tl.load(q_ptrs,
164
- mask=(offs_m[:, None] < seqlen_q) &
165
- (offs_d[None, :] < headdim),
166
- other=0.0)
167
- # loop over k, v and update accumulator
168
- end_n = seqlen_k if not IS_CAUSAL else tl.minimum(
169
- (start_m + 1) * BLOCK_M, seqlen_k)
170
- for start_n in range(0, end_n, BLOCK_N):
171
- start_n = tl.multiple_of(start_n, BLOCK_N)
172
- # -- compute qk ----
173
- if EVEN_N & EVEN_M: # If we just do "if EVEN_N", there seems to be some race condition
174
- if EVEN_HEADDIM:
175
- k = tl.load(k_ptrs + start_n * stride_kn)
176
- else:
177
- k = tl.load(k_ptrs + start_n * stride_kn,
178
- mask=offs_d[None, :] < headdim,
179
- other=0.0)
180
- else:
181
- if EVEN_HEADDIM:
182
- k = tl.load(k_ptrs + start_n * stride_kn,
183
- mask=(start_n + offs_n)[:, None] < seqlen_k,
184
- other=0.0)
185
- else:
186
- k = tl.load(k_ptrs + start_n * stride_kn,
187
- mask=((start_n + offs_n)[:, None] < seqlen_k) &
188
- (offs_d[None, :] < headdim),
189
- other=0.0)
190
- qk = tl.zeros([BLOCK_M, BLOCK_N], dtype=tl.float32)
191
- qk += tl.dot(q, tl.trans(k))
192
- # Trying to combine the two masks seem to make the result wrong
193
- if not EVEN_N: # Need to mask out otherwise the softmax is wrong
194
- qk += tl.where((start_n + offs_n)[None, :] < seqlen_k, 0,
195
- float('-inf'))
196
- if IS_CAUSAL:
197
- qk += tl.where(offs_m[:, None] >= (start_n + offs_n)[None, :], 0,
198
- float('-inf'))
199
- if BIAS_TYPE != 'none':
200
- if BIAS_TYPE == 'vector':
201
- if EVEN_N:
202
- bias = tl.load(b_ptrs + start_n).to(tl.float32)
203
- else:
204
- bias = tl.load(b_ptrs + start_n,
205
- mask=(start_n + offs_n) < seqlen_k,
206
- other=0.0).to(tl.float32)
207
- bias = bias[None, :]
208
- elif BIAS_TYPE == 'matrix':
209
- if EVEN_M & EVEN_N:
210
- bias = tl.load(b_ptrs + start_n).to(tl.float32)
211
- else:
212
- bias = tl.load(b_ptrs + start_n,
213
- mask=(offs_m[:, None] < seqlen_q) &
214
- ((start_n + offs_n)[None, :] < seqlen_k),
215
- other=0.0).to(tl.float32)
216
- else:
217
- raise ValueError(
218
- "BIAS_TYPE must be one of {'vector', 'matrix'}")
219
- # Slightly faster to multiply the softmax_scale in the tl.exp below since the compiler
220
- # can then fuse the mult and add into an fma instruction. But if we have bias we need to
221
- # to multiply with softmax_scale here.
222
- qk = qk * softmax_scale + bias
223
- m_ij = tl.maximum(tl.max(qk, 1), lse_i)
224
- p = tl.exp(qk - m_ij[:, None])
225
- else:
226
- m_ij = tl.maximum(tl.max(qk, 1) * softmax_scale, lse_i)
227
- p = tl.exp(qk * softmax_scale - m_ij[:, None])
228
- l_ij = tl.sum(p, 1)
229
-
230
- # scale acc_o
231
- acc_o_scale = tl.exp(m_i - m_ij)
232
-
233
- # # -- update output accumulator --
234
- # BUG: have to store and immediately load
235
- tl.store(t_ptrs, acc_o_scale)
236
- acc_o_scale = tl.load(t_ptrs)
237
- acc_o = acc_o * acc_o_scale[:, None]
238
- # update acc_o
239
- if EVEN_N & EVEN_M: # If we just do "if EVEN_N", there seems to be some race condition
240
- if EVEN_HEADDIM:
241
- v = tl.load(v_ptrs + start_n * stride_vn)
242
- else:
243
- v = tl.load(v_ptrs + start_n * stride_vn,
244
- mask=offs_d[None, :] < headdim,
245
- other=0.0)
246
- else:
247
- if EVEN_HEADDIM:
248
- v = tl.load(v_ptrs + start_n * stride_vn,
249
- mask=(start_n + offs_n)[:, None] < seqlen_k,
250
- other=0.0)
251
- else:
252
- v = tl.load(v_ptrs + start_n * stride_vn,
253
- mask=((start_n + offs_n)[:, None] < seqlen_k) &
254
- (offs_d[None, :] < headdim),
255
- other=0.0)
256
- p = p.to(v.dtype)
257
- acc_o += tl.dot(p, v)
258
-
259
- # -- update statistics
260
- m_i = m_ij
261
- l_i_new = tl.exp(lse_i - m_ij) + l_ij
262
- lse_i = m_ij + tl.log(l_i_new)
263
-
264
- o_scale = tl.exp(m_i - lse_i)
265
- # BUG: have to store and immediately load
266
- tl.store(t_ptrs, o_scale)
267
- o_scale = tl.load(t_ptrs)
268
- acc_o = acc_o * o_scale[:, None]
269
- # rematerialize offsets to save registers
270
- start_m = tl.program_id(0)
271
- offs_m = start_m * BLOCK_M + tl.arange(0, BLOCK_M)
272
- # write back l and m
273
- lse_ptrs = Lse + off_hb * seqlen_q_rounded + offs_m
274
- tl.store(lse_ptrs, lse_i)
275
- # initialize pointers to output
276
- offs_n = tl.arange(0, BLOCK_HEADDIM)
277
- out_ptrs = Out + off_b * stride_ob + off_h * stride_oh + (
278
- offs_m[:, None] * stride_om + offs_n[None, :])
279
- if EVEN_M:
280
- if EVEN_HEADDIM:
281
- tl.store(out_ptrs, acc_o)
282
- else:
283
- tl.store(out_ptrs, acc_o, mask=offs_d[None, :] < headdim)
284
- else:
285
- if EVEN_HEADDIM:
286
- tl.store(out_ptrs, acc_o, mask=offs_m[:, None] < seqlen_q)
287
- else:
288
- tl.store(out_ptrs,
289
- acc_o,
290
- mask=(offs_m[:, None] < seqlen_q) &
291
- (offs_d[None, :] < headdim))
292
-
293
-
294
- @triton.jit
295
- def _bwd_preprocess_do_o_dot(
296
- Out,
297
- DO,
298
- Delta,
299
- stride_ob,
300
- stride_oh,
301
- stride_om,
302
- stride_dob,
303
- stride_doh,
304
- stride_dom,
305
- nheads,
306
- seqlen_q,
307
- seqlen_q_rounded,
308
- headdim,
309
- BLOCK_M: tl.constexpr,
310
- BLOCK_HEADDIM: tl.constexpr,
311
- ):
312
- start_m = tl.program_id(0)
313
- off_hb = tl.program_id(1)
314
- off_b = off_hb // nheads
315
- off_h = off_hb % nheads
316
- # initialize offsets
317
- offs_m = start_m * BLOCK_M + tl.arange(0, BLOCK_M)
318
- offs_d = tl.arange(0, BLOCK_HEADDIM)
319
- # load
320
- o = tl.load(Out + off_b * stride_ob + off_h * stride_oh +
321
- offs_m[:, None] * stride_om + offs_d[None, :],
322
- mask=(offs_m[:, None] < seqlen_q) & (offs_d[None, :] < headdim),
323
- other=0.0).to(tl.float32)
324
- do = tl.load(DO + off_b * stride_dob + off_h * stride_doh +
325
- offs_m[:, None] * stride_dom + offs_d[None, :],
326
- mask=(offs_m[:, None] < seqlen_q) &
327
- (offs_d[None, :] < headdim),
328
- other=0.0).to(tl.float32)
329
- delta = tl.sum(o * do, axis=1)
330
- # write-back
331
- tl.store(Delta + off_hb * seqlen_q_rounded + offs_m, delta)
332
-
333
-
334
- @triton.jit
335
- def _bwd_kernel_one_col_block(
336
- start_n,
337
- Q,
338
- K,
339
- V,
340
- Bias,
341
- DO,
342
- DQ,
343
- DK,
344
- DV,
345
- LSE,
346
- D,
347
- softmax_scale,
348
- stride_qm,
349
- stride_kn,
350
- stride_vn,
351
- stride_bm,
352
- stride_dom,
353
- stride_dqm,
354
- stride_dkn,
355
- stride_dvn,
356
- seqlen_q,
357
- seqlen_k,
358
- headdim,
359
- ATOMIC_ADD: tl.constexpr,
360
- BIAS_TYPE: tl.constexpr,
361
- IS_CAUSAL: tl.constexpr,
362
- BLOCK_HEADDIM: tl.constexpr,
363
- EVEN_M: tl.constexpr,
364
- EVEN_N: tl.constexpr,
365
- EVEN_HEADDIM: tl.constexpr,
366
- BLOCK_M: tl.constexpr,
367
- BLOCK_N: tl.constexpr,
368
- ):
369
- # We need to make sure begin_m is a multiple of BLOCK_M (not BLOCK_N)
370
- begin_m = 0 if not IS_CAUSAL else ((start_n * BLOCK_N) // BLOCK_M) * BLOCK_M
371
- # initialize row/col offsets
372
- offs_qm = begin_m + tl.arange(0, BLOCK_M)
373
- offs_n = start_n * BLOCK_N + tl.arange(0, BLOCK_N)
374
- offs_m = tl.arange(0, BLOCK_M)
375
- offs_d = tl.arange(0, BLOCK_HEADDIM)
376
- # initialize pointers to value-like data
377
- q_ptrs = Q + (offs_qm[:, None] * stride_qm + offs_d[None, :])
378
- k_ptrs = K + (offs_n[:, None] * stride_kn + offs_d[None, :])
379
- v_ptrs = V + (offs_n[:, None] * stride_vn + offs_d[None, :])
380
- do_ptrs = DO + (offs_qm[:, None] * stride_dom + offs_d[None, :])
381
- dq_ptrs = DQ + (offs_qm[:, None] * stride_dqm + offs_d[None, :])
382
- if BIAS_TYPE == 'vector':
383
- b_ptrs = Bias + offs_n
384
- elif BIAS_TYPE == 'matrix':
385
- b_ptrs = Bias + (offs_qm[:, None] * stride_bm + offs_n[None, :])
386
- else:
387
- raise ValueError("BIAS_TYPE must be one of {'vector', 'matrix'}")
388
- # initialize dv and dk
389
- dv = tl.zeros([BLOCK_N, BLOCK_HEADDIM], dtype=tl.float32)
390
- dk = tl.zeros([BLOCK_N, BLOCK_HEADDIM], dtype=tl.float32)
391
- # k and v stay in SRAM throughout
392
- # [2022-10-30] TD: Same bug as the fwd. In the case of EVEN_N=True and EVEN_M=False,
393
- # if we just call tl.load(k_ptrs), we get the wrong output!
394
- if EVEN_N & EVEN_M:
395
- if EVEN_HEADDIM:
396
- k = tl.load(k_ptrs)
397
- v = tl.load(v_ptrs)
398
- else:
399
- k = tl.load(k_ptrs, mask=offs_d[None, :] < headdim, other=0.0)
400
- v = tl.load(v_ptrs, mask=offs_d[None, :] < headdim, other=0.0)
401
- else:
402
- if EVEN_HEADDIM:
403
- k = tl.load(k_ptrs, mask=offs_n[:, None] < seqlen_k, other=0.0)
404
- v = tl.load(v_ptrs, mask=offs_n[:, None] < seqlen_k, other=0.0)
405
- else:
406
- k = tl.load(k_ptrs,
407
- mask=(offs_n[:, None] < seqlen_k) &
408
- (offs_d[None, :] < headdim),
409
- other=0.0)
410
- v = tl.load(v_ptrs,
411
- mask=(offs_n[:, None] < seqlen_k) &
412
- (offs_d[None, :] < headdim),
413
- other=0.0)
414
- # loop over rows
415
- num_block_m = tl.cdiv(seqlen_q, BLOCK_M)
416
- for start_m in range(begin_m, num_block_m * BLOCK_M, BLOCK_M):
417
- start_m = tl.multiple_of(start_m, BLOCK_M)
418
- offs_m_curr = start_m + offs_m
419
- # load q, k, v, do on-chip
420
- # Same bug as below. Otherwise gives wrong result for headdim=40, seqlen=(128, 117)
421
- if EVEN_M & EVEN_HEADDIM:
422
- q = tl.load(q_ptrs)
423
- else:
424
- if EVEN_HEADDIM:
425
- q = tl.load(q_ptrs,
426
- mask=offs_m_curr[:, None] < seqlen_q,
427
- other=0.0)
428
- else:
429
- q = tl.load(q_ptrs,
430
- mask=(offs_m_curr[:, None] < seqlen_q) &
431
- (offs_d[None, :] < headdim),
432
- other=0.0)
433
- # recompute p = softmax(qk, dim=-1).T
434
- qk = tl.dot(q, tl.trans(k))
435
- # Trying to combine the two masks seem to make the result wrong
436
- if not EVEN_N: # Need to mask out otherwise the softmax is wrong
437
- qk = tl.where(offs_n[None, :] < seqlen_k, qk, float('-inf'))
438
- if IS_CAUSAL:
439
- qk = tl.where(offs_m_curr[:, None] >= (offs_n[None, :]), qk,
440
- float('-inf'))
441
- if BIAS_TYPE != 'none':
442
- if BIAS_TYPE == 'vector':
443
- if EVEN_N:
444
- bias = tl.load(b_ptrs).to(tl.float32)
445
- else:
446
- bias = tl.load(b_ptrs, mask=offs_n < seqlen_k,
447
- other=0.0).to(tl.float32)
448
- bias = bias[None, :]
449
- elif BIAS_TYPE == 'matrix':
450
- if EVEN_M & EVEN_N:
451
- bias = tl.load(b_ptrs).to(tl.float32)
452
- else:
453
- bias = tl.load(b_ptrs,
454
- mask=(offs_m_curr[:, None] < seqlen_q) &
455
- (offs_n[None, :] < seqlen_k),
456
- other=0.0).to(tl.float32)
457
- else:
458
- raise ValueError(
459
- "BIAS_TYPE must be one of {'vector', 'matrix'}")
460
- qk = qk * softmax_scale + bias
461
- # There seems to be a race condition when headdim=48/96, and dq, dk, dv are wrong.
462
- # Also wrong for headdim=64.
463
- if not (EVEN_M & EVEN_HEADDIM):
464
- tl.debug_barrier()
465
- lse_i = tl.load(LSE + offs_m_curr)
466
- if BIAS_TYPE == 'none':
467
- p = tl.exp(qk * softmax_scale - lse_i[:, None])
468
- else:
469
- p = tl.exp(qk - lse_i[:, None])
470
- # compute dv
471
- # [2022-10-30] TD: A Triton bug: if EVEN_M=True and EVEN_HEADDIM=False, if we call
472
- # do = tl.load(do_ptrs, mask=offs_d[None, :] < headdim, other=0.0), we get wrong outputs
473
- # in the case of headdim=48/96, seqlen_q & seqlen_k >= 512. If headdim=40 or seqlen < 512,
474
- # the output is correct.
475
- if EVEN_M & EVEN_HEADDIM:
476
- do = tl.load(do_ptrs)
477
- else:
478
- # [2022-11-01] TD: Triton bug, there's a race condition if we just use m_mask and not d_mask.
479
- do = tl.load(do_ptrs,
480
- mask=(offs_m_curr[:, None] < seqlen_q) &
481
- (offs_d[None, :] < headdim),
482
- other=0.0)
483
- # if EVEN_M:
484
- # if EVEN_HEADDIM:
485
- # do = tl.load(do_ptrs)
486
- # else:
487
- # do = tl.load(do_ptrs, mask=offs_d[None, :] < headdim, other=0.0)
488
- # else:
489
- # if EVEN_HEADDIM:
490
- # do = tl.load(do_ptrs, mask=offs_m_curr[:, None] < seqlen_q, other=0.0)
491
- # else:
492
- # do = tl.load(do_ptrs, mask=(offs_m_curr[:, None] < seqlen_q)
493
- # & (offs_d[None, :] < headdim), other=0.0)
494
- dv += tl.dot(tl.trans(p).to(do.dtype), do)
495
- # compute dp = dot(v, do)
496
- # There seems to be a race condition when headdim=48/96, and dq, dk are wrong.
497
- # Also wrong for headdim=128, seqlen=(108, 256), and ATOMIC_ADD=True
498
- # Also wrong for headdim=64, seqlen=(1023, 1024), and ATOMIC_ADD=False
499
- if not (EVEN_M & EVEN_HEADDIM):
500
- tl.debug_barrier()
501
- dp = tl.dot(do, v, trans_b=True)
502
- # There's a race condition for headdim=48
503
- if not EVEN_HEADDIM:
504
- tl.debug_barrier()
505
- # compute ds = p * (dp - delta[:, None])
506
- # Putting the subtraction after the dp matmul (instead of before) is slightly faster
507
- Di = tl.load(D + offs_m_curr)
508
- # Converting ds to q.dtype here reduces register pressure and makes it much faster
509
- # for BLOCK_HEADDIM=128
510
- ds = (p * (dp - Di[:, None]) * softmax_scale).to(q.dtype)
511
- # compute dk = dot(ds.T, q)
512
- dk += tl.dot(tl.trans(ds), q)
513
- # compute dq
514
- if not ATOMIC_ADD:
515
- if EVEN_M & EVEN_HEADDIM: # Race condition if we just do EVEN_M
516
- dq = tl.load(dq_ptrs, eviction_policy='evict_last')
517
- dq += tl.dot(ds, k)
518
- tl.store(dq_ptrs, dq, eviction_policy='evict_last')
519
- else:
520
- if EVEN_HEADDIM:
521
- dq = tl.load(dq_ptrs,
522
- mask=offs_m_curr[:, None] < seqlen_q,
523
- other=0.0,
524
- eviction_policy='evict_last')
525
- dq += tl.dot(ds, k)
526
- tl.store(dq_ptrs,
527
- dq,
528
- mask=offs_m_curr[:, None] < seqlen_q,
529
- eviction_policy='evict_last')
530
- else:
531
- dq = tl.load(dq_ptrs,
532
- mask=(offs_m_curr[:, None] < seqlen_q) &
533
- (offs_d[None, :] < headdim),
534
- other=0.0,
535
- eviction_policy='evict_last')
536
- dq += tl.dot(ds, k)
537
- tl.store(dq_ptrs,
538
- dq,
539
- mask=(offs_m_curr[:, None] < seqlen_q) &
540
- (offs_d[None, :] < headdim),
541
- eviction_policy='evict_last')
542
- else: # If we're parallelizing across the seqlen_k dimension
543
- dq = tl.dot(ds, k)
544
- if EVEN_M & EVEN_HEADDIM: # Race condition if we just do EVEN_M
545
- tl.atomic_add(dq_ptrs, dq)
546
- else:
547
- if EVEN_HEADDIM:
548
- tl.atomic_add(dq_ptrs,
549
- dq,
550
- mask=offs_m_curr[:, None] < seqlen_q)
551
- else:
552
- tl.atomic_add(dq_ptrs,
553
- dq,
554
- mask=(offs_m_curr[:, None] < seqlen_q) &
555
- (offs_d[None, :] < headdim))
556
- # increment pointers
557
- dq_ptrs += BLOCK_M * stride_dqm
558
- q_ptrs += BLOCK_M * stride_qm
559
- do_ptrs += BLOCK_M * stride_dom
560
- if BIAS_TYPE == 'matrix':
561
- b_ptrs += BLOCK_M * stride_bm
562
- # write-back
563
- dv_ptrs = DV + (offs_n[:, None] * stride_dvn + offs_d[None, :])
564
- dk_ptrs = DK + (offs_n[:, None] * stride_dkn + offs_d[None, :])
565
- # [2022-11-01] TD: Same bug. In the case of EVEN_N=True and EVEN_M=False,
566
- # if we just call tl.store(dv_ptrs), there's a race condition
567
- if EVEN_N & EVEN_M:
568
- if EVEN_HEADDIM:
569
- tl.store(dv_ptrs, dv)
570
- tl.store(dk_ptrs, dk)
571
- else:
572
- tl.store(dv_ptrs, dv, mask=offs_d[None, :] < headdim)
573
- tl.store(dk_ptrs, dk, mask=offs_d[None, :] < headdim)
574
- else:
575
- if EVEN_HEADDIM:
576
- tl.store(dv_ptrs, dv, mask=offs_n[:, None] < seqlen_k)
577
- tl.store(dk_ptrs, dk, mask=offs_n[:, None] < seqlen_k)
578
- else:
579
- tl.store(dv_ptrs,
580
- dv,
581
- mask=(offs_n[:, None] < seqlen_k) &
582
- (offs_d[None, :] < headdim))
583
- tl.store(dk_ptrs,
584
- dk,
585
- mask=(offs_n[:, None] < seqlen_k) &
586
- (offs_d[None, :] < headdim))
587
-
588
-
589
- def init_to_zero(name):
590
- return lambda nargs: nargs[name].zero_()
591
-
592
-
593
- @triton.autotune(
594
- configs=[
595
- triton.Config(
596
- {
597
- 'BLOCK_M': 128,
598
- 'BLOCK_N': 128,
599
- 'SEQUENCE_PARALLEL': False
600
- },
601
- num_warps=8,
602
- num_stages=1,
603
- pre_hook=init_to_zero('DQ')),
604
- triton.Config(
605
- {
606
- 'BLOCK_M': 128,
607
- 'BLOCK_N': 128,
608
- 'SEQUENCE_PARALLEL': True
609
- },
610
- num_warps=8,
611
- num_stages=1,
612
- pre_hook=init_to_zero('DQ')),
613
- # Other configs seem to give wrong results when seqlen_q % 128 != 0, disabling them for now
614
- # # Kernel is buggy (give wrong result) if we set BLOCK_m=128, BLOCK_n=64, num_warps=*4*
615
- # triton.Config({"BLOCK_M": 128, "BLOCK_N": 64, "SEQUENCE_PARALLEL": False}, num_warps=8, num_stages=1, pre_hook=init_to_zero('DQ')),
616
- # triton.Config({"BLOCK_M": 128, "BLOCK_N": 64, "SEQUENCE_PARALLEL": True}, num_warps=8, num_stages=1, pre_hook=init_to_zero('DQ')),
617
- # triton.Config({"BLOCK_M": 64, "BLOCK_N": 64, "SEQUENCE_PARALLEL": False}, num_warps=4, num_stages=1, pre_hook=init_to_zero('DQ')),
618
- # triton.Config({"BLOCK_M": 64, "BLOCK_N": 64, "SEQUENCE_PARALLEL": True}, num_warps=4, num_stages=1, pre_hook=init_to_zero('DQ')),
619
- ],
620
- key=[
621
- 'CACHE_KEY_SEQLEN_Q', 'CACHE_KEY_SEQLEN_K', 'BIAS_TYPE', 'IS_CAUSAL',
622
- 'BLOCK_HEADDIM'
623
- ],
624
- )
625
- @triton.heuristics({
626
- 'EVEN_M': lambda args: args['seqlen_q'] % args['BLOCK_M'] == 0,
627
- 'EVEN_N': lambda args: args['seqlen_k'] % args['BLOCK_N'] == 0,
628
- 'EVEN_HEADDIM': lambda args: args['headdim'] == args['BLOCK_HEADDIM'],
629
- })
630
- @triton.jit
631
- def _bwd_kernel(
632
- Q,
633
- K,
634
- V,
635
- Bias,
636
- DO,
637
- DQ,
638
- DK,
639
- DV,
640
- LSE,
641
- D,
642
- softmax_scale,
643
- stride_qb,
644
- stride_qh,
645
- stride_qm,
646
- stride_kb,
647
- stride_kh,
648
- stride_kn,
649
- stride_vb,
650
- stride_vh,
651
- stride_vn,
652
- stride_bb,
653
- stride_bh,
654
- stride_bm,
655
- stride_dob,
656
- stride_doh,
657
- stride_dom,
658
- stride_dqb,
659
- stride_dqh,
660
- stride_dqm,
661
- stride_dkb,
662
- stride_dkh,
663
- stride_dkn,
664
- stride_dvb,
665
- stride_dvh,
666
- stride_dvn,
667
- nheads,
668
- seqlen_q,
669
- seqlen_k,
670
- seqlen_q_rounded,
671
- headdim,
672
- CACHE_KEY_SEQLEN_Q,
673
- CACHE_KEY_SEQLEN_K,
674
- BIAS_TYPE: tl.constexpr,
675
- IS_CAUSAL: tl.constexpr,
676
- BLOCK_HEADDIM: tl.constexpr,
677
- SEQUENCE_PARALLEL: tl.constexpr,
678
- EVEN_M: tl.constexpr,
679
- EVEN_N: tl.constexpr,
680
- EVEN_HEADDIM: tl.constexpr,
681
- BLOCK_M: tl.constexpr,
682
- BLOCK_N: tl.constexpr,
683
- ):
684
- off_hb = tl.program_id(1)
685
- off_b = off_hb // nheads
686
- off_h = off_hb % nheads
687
- # offset pointers for batch/head
688
- Q += off_b * stride_qb + off_h * stride_qh
689
- K += off_b * stride_kb + off_h * stride_kh
690
- V += off_b * stride_vb + off_h * stride_vh
691
- DO += off_b * stride_dob + off_h * stride_doh
692
- DQ += off_b * stride_dqb + off_h * stride_dqh
693
- DK += off_b * stride_dkb + off_h * stride_dkh
694
- DV += off_b * stride_dvb + off_h * stride_dvh
695
- if BIAS_TYPE != 'none':
696
- Bias += off_b * stride_bb + off_h * stride_bh
697
- # pointer to row-wise quantities in value-like data
698
- D += off_hb * seqlen_q_rounded
699
- LSE += off_hb * seqlen_q_rounded
700
- if not SEQUENCE_PARALLEL:
701
- num_block_n = tl.cdiv(seqlen_k, BLOCK_N)
702
- for start_n in range(0, num_block_n):
703
- _bwd_kernel_one_col_block(start_n,
704
- Q,
705
- K,
706
- V,
707
- Bias,
708
- DO,
709
- DQ,
710
- DK,
711
- DV,
712
- LSE,
713
- D,
714
- softmax_scale,
715
- stride_qm,
716
- stride_kn,
717
- stride_vn,
718
- stride_bm,
719
- stride_dom,
720
- stride_dqm,
721
- stride_dkn,
722
- stride_dvn,
723
- seqlen_q,
724
- seqlen_k,
725
- headdim,
726
- ATOMIC_ADD=False,
727
- BIAS_TYPE=BIAS_TYPE,
728
- IS_CAUSAL=IS_CAUSAL,
729
- BLOCK_HEADDIM=BLOCK_HEADDIM,
730
- EVEN_M=EVEN_M,
731
- EVEN_N=EVEN_N,
732
- EVEN_HEADDIM=EVEN_HEADDIM,
733
- BLOCK_M=BLOCK_M,
734
- BLOCK_N=BLOCK_N)
735
- else:
736
- start_n = tl.program_id(0)
737
- _bwd_kernel_one_col_block(start_n,
738
- Q,
739
- K,
740
- V,
741
- Bias,
742
- DO,
743
- DQ,
744
- DK,
745
- DV,
746
- LSE,
747
- D,
748
- softmax_scale,
749
- stride_qm,
750
- stride_kn,
751
- stride_vn,
752
- stride_bm,
753
- stride_dom,
754
- stride_dqm,
755
- stride_dkn,
756
- stride_dvn,
757
- seqlen_q,
758
- seqlen_k,
759
- headdim,
760
- ATOMIC_ADD=True,
761
- BIAS_TYPE=BIAS_TYPE,
762
- IS_CAUSAL=IS_CAUSAL,
763
- BLOCK_HEADDIM=BLOCK_HEADDIM,
764
- EVEN_M=EVEN_M,
765
- EVEN_N=EVEN_N,
766
- EVEN_HEADDIM=EVEN_HEADDIM,
767
- BLOCK_M=BLOCK_M,
768
- BLOCK_N=BLOCK_N)
769
-
770
-
771
- def _flash_attn_forward(q, k, v, bias=None, causal=False, softmax_scale=None):
772
- # shape constraints
773
- batch, seqlen_q, nheads, d = q.shape
774
- _, seqlen_k, _, _ = k.shape
775
- assert k.shape == (batch, seqlen_k, nheads, d)
776
- assert v.shape == (batch, seqlen_k, nheads, d)
777
- assert d <= 128, 'FlashAttention only support head dimensions up to 128'
778
- assert q.dtype == k.dtype == v.dtype, 'All tensors must have the same type'
779
- assert q.dtype in [torch.float16,
780
- torch.bfloat16], 'Only support fp16 and bf16'
781
- assert q.is_cuda and k.is_cuda and v.is_cuda
782
- softmax_scale = softmax_scale or 1.0 / math.sqrt(d)
783
-
784
- has_bias = bias is not None
785
- bias_type = 'none'
786
- if has_bias:
787
- assert bias.dtype in [q.dtype, torch.float]
788
- assert bias.is_cuda
789
- assert bias.dim() == 4
790
- if bias.stride(-1) != 1:
791
- bias = bias.contiguous()
792
- if bias.shape[2:] == (1, seqlen_k):
793
- bias_type = 'vector'
794
- elif bias.shape[2:] == (seqlen_q, seqlen_k):
795
- bias_type = 'matrix'
796
- else:
797
- raise RuntimeError('Last 2 dimensions of bias must be (1, seqlen_k)'
798
- ' or (seqlen_q, seqlen_k)')
799
- if bias.shape[:2] == (1, nheads):
800
- bias = repeat(bias, '1 h ... -> b h ...', b=batch)
801
- elif bias.shape[:2] == (batch, 1):
802
- bias = repeat(bias, 'b 1 ... -> b h ...', h=nheads)
803
- elif bias.shape[:2] == (1, 1):
804
- bias = repeat(bias, '1 h ... -> b h ...', b=batch)
805
- bias = repeat(bias, 'b 1 ... -> b h ...', h=nheads)
806
- assert bias.shape[:2] == (
807
- batch, nheads
808
- ), f'First 2 dimensions of bias must be broadcastible to (batch, nheads) = ({batch, nheads}). Bias has shape: {bias.shape}'
809
- assert bias is not None # for type checking
810
- bias_strides = (bias.stride(0), bias.stride(1),
811
- bias.stride(2)) if has_bias else (0, 0, 0)
812
-
813
- seqlen_q_rounded = math.ceil(seqlen_q / 128) * 128
814
- lse = torch.empty((batch, nheads, seqlen_q_rounded),
815
- device=q.device,
816
- dtype=torch.float32)
817
- tmp = torch.empty((batch, nheads, seqlen_q_rounded),
818
- device=q.device,
819
- dtype=torch.float32)
820
- o = torch.empty_like(q)
821
-
822
- BLOCK_HEADDIM = max(triton.next_power_of_2(d), 16)
823
- # BLOCK = 128
824
- # num_warps = 4 if d <= 64 else 8
825
- grid = lambda META: (triton.cdiv(seqlen_q, META['BLOCK_M']), batch * nheads)
826
- _fwd_kernel[grid]( # type: ignore
827
- q,
828
- k,
829
- v,
830
- bias,
831
- o,
832
- lse,
833
- tmp,
834
- softmax_scale,
835
- q.stride(0),
836
- q.stride(2),
837
- q.stride(1),
838
- k.stride(0),
839
- k.stride(2),
840
- k.stride(1),
841
- v.stride(0),
842
- v.stride(2),
843
- v.stride(1),
844
- *bias_strides,
845
- o.stride(0),
846
- o.stride(2),
847
- o.stride(1),
848
- nheads,
849
- seqlen_q,
850
- seqlen_k,
851
- seqlen_q_rounded,
852
- d,
853
- seqlen_q // 32,
854
- seqlen_k // 32, # key for triton cache (limit number of compilations)
855
- # Can't use kwargs here because triton autotune expects key to be args, not kwargs
856
- # IS_CAUSAL=causal, BLOCK_HEADDIM=d,
857
- bias_type,
858
- causal,
859
- BLOCK_HEADDIM,
860
- # BLOCK_M=BLOCK, BLOCK_N=BLOCK,
861
- # num_warps=num_warps,
862
- # num_stages=1,
863
- )
864
- return o, lse, softmax_scale # softmax_scale could have been updated
865
-
866
-
867
- def _flash_attn_backward(do,
868
- q,
869
- k,
870
- v,
871
- o,
872
- lse,
873
- dq,
874
- dk,
875
- dv,
876
- bias=None,
877
- causal=False,
878
- softmax_scale=None):
879
- # Make sure that the last dimension is contiguous
880
- if do.stride(-1) != 1:
881
- do = do.contiguous()
882
- batch, seqlen_q, nheads, d = q.shape
883
- _, seqlen_k, _, _ = k.shape
884
- # assert d in {16, 32, 64, 128}
885
- assert d <= 128
886
- seqlen_q_rounded = math.ceil(seqlen_q / 128) * 128
887
- assert lse.shape == (batch, nheads, seqlen_q_rounded)
888
- assert q.stride(-1) == k.stride(-1) == v.stride(-1) == o.stride(-1) == 1
889
- assert dq.stride(-1) == dk.stride(-1) == dv.stride(-1) == 1
890
- softmax_scale = softmax_scale or 1.0 / math.sqrt(d)
891
- # dq_accum = torch.zeros_like(q, dtype=torch.float32)
892
- dq_accum = torch.empty_like(q, dtype=torch.float32)
893
- delta = torch.empty_like(lse)
894
- # delta = torch.zeros_like(lse)
895
-
896
- BLOCK_HEADDIM = max(triton.next_power_of_2(d), 16)
897
- grid = lambda META: (triton.cdiv(seqlen_q, META['BLOCK_M']), batch * nheads)
898
- _bwd_preprocess_do_o_dot[grid]( # type: ignore
899
- o,
900
- do,
901
- delta,
902
- o.stride(0),
903
- o.stride(2),
904
- o.stride(1),
905
- do.stride(0),
906
- do.stride(2),
907
- do.stride(1),
908
- nheads,
909
- seqlen_q,
910
- seqlen_q_rounded,
911
- d,
912
- BLOCK_M=128,
913
- BLOCK_HEADDIM=BLOCK_HEADDIM,
914
- )
915
-
916
- has_bias = bias is not None
917
- bias_type = 'none'
918
- if has_bias:
919
- assert bias.dtype in [q.dtype, torch.float]
920
- assert bias.is_cuda
921
- assert bias.dim() == 4
922
- assert bias.stride(-1) == 1
923
- if bias.shape[2:] == (1, seqlen_k):
924
- bias_type = 'vector'
925
- elif bias.shape[2:] == (seqlen_q, seqlen_k):
926
- bias_type = 'matrix'
927
- else:
928
- raise RuntimeError('Last 2 dimensions of bias must be (1, seqlen_k)'
929
- ' or (seqlen_q, seqlen_k)')
930
- if bias.shape[:2] == (1, nheads):
931
- bias = repeat(bias, '1 h ... -> b h ...', b=batch)
932
- elif bias.shape[:2] == (batch, 1):
933
- bias = repeat(bias, 'b 1 ... -> b h ...', h=nheads)
934
- elif bias.shape[:2] == (1, 1):
935
- bias = repeat(bias, '1 h ... -> b h ...', b=batch)
936
- bias = repeat(bias, 'b 1 ... -> b h ...', h=nheads)
937
- assert bias.shape[:2] == (
938
- batch, nheads
939
- ), f'First 2 dimensions of bias must be broadcastible to (batch, nheads) = ({batch, nheads}). Bias has shape: {bias.shape}'
940
- assert bias is not None # type checking
941
- bias_strides = (bias.stride(0), bias.stride(1),
942
- bias.stride(2)) if has_bias else (0, 0, 0)
943
-
944
- # BLOCK_M = 128
945
- # BLOCK_N = 64
946
- # num_warps = 4
947
- grid = lambda META: (triton.cdiv(seqlen_k, META['BLOCK_N'])
948
- if META['SEQUENCE_PARALLEL'] else 1, batch * nheads)
949
- _bwd_kernel[grid]( # type: ignore
950
- q,
951
- k,
952
- v,
953
- bias,
954
- do,
955
- dq_accum,
956
- dk,
957
- dv,
958
- lse,
959
- delta,
960
- softmax_scale,
961
- q.stride(0),
962
- q.stride(2),
963
- q.stride(1),
964
- k.stride(0),
965
- k.stride(2),
966
- k.stride(1),
967
- v.stride(0),
968
- v.stride(2),
969
- v.stride(1),
970
- *bias_strides,
971
- do.stride(0),
972
- do.stride(2),
973
- do.stride(1),
974
- dq_accum.stride(0),
975
- dq_accum.stride(2),
976
- dq_accum.stride(1),
977
- dk.stride(0),
978
- dk.stride(2),
979
- dk.stride(1),
980
- dv.stride(0),
981
- dv.stride(2),
982
- dv.stride(1),
983
- nheads,
984
- seqlen_q,
985
- seqlen_k,
986
- seqlen_q_rounded,
987
- d,
988
- seqlen_q // 32,
989
- seqlen_k // 32, # key for triton cache (limit number of compilations)
990
- # Can't use kwargs here because triton autotune expects key to be args, not kwargs
991
- # IS_CAUSAL=causal, BLOCK_HEADDIM=d,
992
- bias_type,
993
- causal,
994
- BLOCK_HEADDIM,
995
- # SEQUENCE_PARALLEL=False,
996
- # BLOCK_M=BLOCK_M, BLOCK_N=BLOCK_N,
997
- # num_warps=num_warps,
998
- # num_stages=1,
999
- )
1000
- dq.copy_(dq_accum)
1001
-
1002
-
1003
- class _FlashAttnQKVPackedFunc(torch.autograd.Function):
1004
-
1005
- @staticmethod
1006
- def forward(ctx, qkv, bias=None, causal=False, softmax_scale=None):
1007
- """Forward pass for packed FlashAttention.
1008
-
1009
- Args:
1010
- ctx: autograd context
1011
- qkv: (batch, seqlen, 3, nheads, headdim)
1012
- bias: optional, shape broadcastible to (batch, nheads, seqlen, seqlen).
1013
- For example, ALiBi mask for causal would have shape (1, nheads, 1, seqlen).
1014
- ALiBi mask for non-causal would have shape (1, nheads, seqlen, seqlen)
1015
- causal (bool): whether to incorporate causal attention masking
1016
- softmax_scale (float, optional): scale factor for softmax
1017
- """
1018
- # Make sure that the last dimension is contiguous
1019
- if qkv.stride(-1) != 1:
1020
- qkv = qkv.contiguous()
1021
- o, lse, ctx.softmax_scale = _flash_attn_forward(
1022
- qkv[:, :, 0],
1023
- qkv[:, :, 1],
1024
- qkv[:, :, 2],
1025
- bias=bias,
1026
- causal=causal,
1027
- softmax_scale=softmax_scale)
1028
- ctx.save_for_backward(qkv, o, lse, bias)
1029
- ctx.causal = causal
1030
- return o
1031
-
1032
- @staticmethod
1033
- def backward(ctx, do):
1034
- qkv, o, lse, bias = ctx.saved_tensors
1035
- assert not ctx.needs_input_grad[
1036
- 1], 'FlashAttention does not support bias gradient yet'
1037
- # Triton's autotune causes the Tensor._version to change, and so Pytorch autograd
1038
- # does a memcpy. To avoid this we run in inference_mode, which doesn't track the version.
1039
- with torch.inference_mode():
1040
- dqkv = torch.empty_like(qkv)
1041
- _flash_attn_backward(do,
1042
- qkv[:, :, 0],
1043
- qkv[:, :, 1],
1044
- qkv[:, :, 2],
1045
- o,
1046
- lse,
1047
- dqkv[:, :, 0],
1048
- dqkv[:, :, 1],
1049
- dqkv[:, :, 2],
1050
- bias=bias,
1051
- causal=ctx.causal,
1052
- softmax_scale=ctx.softmax_scale)
1053
- return dqkv, None, None, None
1054
-
1055
-
1056
- flash_attn_qkvpacked_func = _FlashAttnQKVPackedFunc.apply
1057
-
1058
-
1059
- class _FlashAttnFunc(torch.autograd.Function):
1060
-
1061
- @staticmethod
1062
- def forward(ctx, q, k, v, bias=None, causal=False, softmax_scale=None):
1063
- """Forward pass for FlashAttention.
1064
-
1065
- Args:
1066
- ctx: autograd context
1067
- q: (batch_size, seqlen_q, nheads, headdim)
1068
- k: (batch_size, seqlen_k, nheads, headdim)
1069
- v: (batch_size, seqlen_k, nheads, headdim)
1070
- bias: optional, shape broadcastible to (batch, nheads, seqlen_q, seqlen_k).
1071
- For example, ALiBi mask for causal would have shape (1, nheads, 1, seqlen_k).
1072
- ALiBi mask for non-causal would have shape (1, nheads, seqlen_q, seqlen_k)
1073
- causal (bool): whether to incorporate causal attention masking
1074
- softmax_scale (float, optional): scale factor for softmax
1075
- """
1076
- # Make sure that the last dimension is contiguous
1077
- q, k, v = [
1078
- x if x.stride(-1) == 1 else x.contiguous() for x in [q, k, v]
1079
- ]
1080
- o, lse, ctx.softmax_scale = _flash_attn_forward(
1081
- q, k, v, bias=bias, causal=causal, softmax_scale=softmax_scale)
1082
- ctx.save_for_backward(q, k, v, o, lse, bias)
1083
- ctx.causal = causal
1084
- return o
1085
-
1086
- @staticmethod
1087
- def backward(ctx, do):
1088
- q, k, v, o, lse, bias = ctx.saved_tensors
1089
- assert not ctx.needs_input_grad[
1090
- 3], 'FlashAttention does not support bias gradient yet'
1091
- # Triton's autotune causes the Tensor._version to change, and so Pytorch autograd
1092
- # does a memcpy. To avoid this we run in inference_mode, which doesn't track the version.
1093
- with torch.inference_mode():
1094
- dq = torch.empty_like(q)
1095
- dk = torch.empty_like(k)
1096
- dv = torch.empty_like(v)
1097
- _flash_attn_backward(do,
1098
- q,
1099
- k,
1100
- v,
1101
- o,
1102
- lse,
1103
- dq,
1104
- dk,
1105
- dv,
1106
- bias=bias,
1107
- causal=ctx.causal,
1108
- softmax_scale=ctx.softmax_scale)
1109
- return dq, dk, dv, None, None, None
1110
-
1111
-
1112
- flash_attn_func = _FlashAttnFunc.apply